using System.Timers;
using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using GraphicsManager.Objects.Core;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using Timer = System.Timers.Timer;

namespace GraphicsManager.Objects;

public class FlowLayout : IRenderObject, IParent
{
    private Rectangle _bounds;

	public FlowLayout()
	{
		_bounds = new Rectangle();
		_bounds.Clicked += _bounds_Clicked;
		t = new(50);
		t.Enabled = true;
		t.Elapsed += TOnElapsed;
		t.Start();
	}

	private Task _bounds_Clicked(IRenderObject arg)
	{
		if (Clicked is not null) _ = Clicked.Invoke(arg);
		return Task.CompletedTask;
	}

	public ControlList Controls { get; } = new();
	public ObjectAnchor Anchor { get => _bounds.Anchor; set => _bounds.Anchor = value; }
	public Color4 BackgroundColor
	{
		get
		{
			Uniform<Vector4>? u4 = _bounds.Uniforms.Uniform4.Where(u => u.Location == 0).First();
			if (u4 is null) u4 = new() { Location = 0, Value = new(1, 1, 0, 1) };
			return new Color4(u4.Value.X, u4.Value.X, u4.Value.X, u4.Value.X);
		}
		set
		{
			Uniform<Vector4> u4 = _bounds.Uniforms.Uniform4.Where(u => u.Location == 0).First();
			if (u4 is not null) _bounds.Uniforms.Uniform4.Remove(u4);
			if (u4 is null) u4 = new() { Location = 0 };
			u4.Value = new(value.R, value.G, value.B, value.A);
			_bounds.Uniforms.Uniform4.Add(u4);
		}
	}
	public bool Visible { get => _bounds.Visible; set => _bounds.Visible = value; }
	public Vector2i Size { get => _bounds.Size; set => _bounds.Size = value; }
	public Vector2 SizeAsFloat { get => _bounds.SizeAsFloat; }
	public Vector2i Location { get => _bounds.Location; set => _bounds.Location = value; }
	public Vector2i Position => Location;
	public Vector2 LocationAsFloat { get => _bounds.LocationAsFloat; }
	public Vector2i Distance { get => _bounds.Distance; }
	public event Func<IRenderObject, Task>? Clicked;
	public event Action<MouseButtonEventArgs> MouseDown;
	public event Action<KeyboardKeyEventArgs> KeyDown;
	public event Func<IRenderObject, Task>? WindowLoaded;
	public event Func<IRenderObject, Task>? MouseEnter;
	public event Func<IRenderObject, Task>? MouseLeave;
	public object? Tag { get; set; } = null;
	public IParent? Parent { get; private set; }
	public Window? Window { get; private set; }
	public bool Loaded { get; private set; } = false;

	public Vector2 MousePosition => Window!.MousePosition;

	public void LoadToParent(IParent Parent, Window Window)
	{
		if (Loaded) return;
		this.Parent = Parent;
		this.Window = Window;
		Loaded = true;
		_bounds.LoadToParent(Parent, Window);
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].LoadToParent(this, Window);
		}
		Window.MouseWheel += WindowOnMouseWheel;
		if (WindowLoaded is not null) WindowLoaded.Invoke(this);
	}

	private bool isrole = false;
	private MouseWheelEventArgs dis;
	Timer t;
	private Queue<Action> scrols = new();
	
	public void TryDraw()
	{
		if (!isrole) Parent!.TryDraw();
	}
	private void WindowOnMouseWheel(MouseWheelEventArgs obj)
	{
		try
		{
			isrole = true;
			dis = obj;
			if (scrols.Any())
			{
				for (int i = 0; i < scrols.Count; i++)
				{
					scrols.Dequeue().Invoke();
				}
				Parent!.TryDraw();
			}
		}
		catch (Exception e)
		{
			Console.WriteLine(e);
		}
		
	}

	public uint HScrollPixels { get; set; } = 30;
	private void TOnElapsed(object? sender, ElapsedEventArgs e)
	{
		try
		{
			if (isrole)
			{
				scrols.Enqueue(new Action(() =>
				{
					for (int i = 0; i < Controls.Length; i++)
					{
						Controls[i].Location = new((int)(Controls[i].Location.X),
							(int)(Controls[i].Location.Y - (dis.OffsetY * HScrollPixels)));
					}
				}));
			}

			isrole = false;
		}
		catch (Exception exception)
		{
			Console.WriteLine(exception);
		}
	}

	public void Draw()
	{
		if (Loaded)
		{
			_bounds.Draw();
			IEnumerable<IRenderObject> needload = Controls.Where(a => a.Loaded == false);

			if (needload.Any())
			{
				foreach (IRenderObject Control in needload)
				{
					Control.LoadToParent(this, Window!);
				}
			}
			for (int i = 0; i < Controls.Length; i++)
			{
				Controls[i].Draw();
			}
		}
	}

	public void Clean()
	{
		for (int i = 0; i < Controls.Length; i++)
		{
			Controls[i].Clean();
		}
		_bounds.Clean();
	}

	public void ParentResize(ResizeEventArgs e)
	{
		if (e.Width == 0 && e.Height == 0) return;
		for (int i = 0; i < Controls.Length; i++)
		{
			if (!Controls[i].Loaded) continue;
			bool top = (Controls[i].Anchor & ObjectAnchor.Top) == ObjectAnchor.Top;
			bool left = (Controls[i].Anchor & ObjectAnchor.Left) == ObjectAnchor.Left;
			bool right = (Controls[i].Anchor & ObjectAnchor.Right) == ObjectAnchor.Right;
			bool bottom = (Controls[i].Anchor & ObjectAnchor.Bottom) == ObjectAnchor.Bottom;
			if (!top && !bottom) { Controls[i].Anchor |= ObjectAnchor.Top; top = true; }
			if (!left && !right) { Controls[i].Anchor |= ObjectAnchor.Left; left = true; }
			int lx = (left ? Controls[i].Location.X : Size.X - Controls[i].Distance.X - Controls[i].Size.X);
			int ly = (top ? Controls[i].Location.Y : Size.Y - Controls[i].Distance.Y - Controls[i].Size.Y);
			int sy = (bottom ? Size.Y - Controls[i].Distance.Y - ly : Controls[i].Size.Y);
			int sx = (right ? Size.X - Controls[i].Distance.X - lx : Controls[i].Size.X);
			Controls[i].Size = new(sx, sy);
			Controls[i].Location = new(lx, ly);
			if (Controls[i] is IParent parent)
			{
				parent.ParentResize(e);
			}
		}
	}

	#region Cool Math Things
	public float[] RctToFloat(int x, int y, int Width, int Height, bool hastexture = false, float z = 0.0f)
	{
		if (hastexture)
		{
			return new float[20] {
				IntToFloat(x + Width), IntToFloat(y, true), z, 1.0f, 1.0f,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, 1.0f, 0.0f,//b r
                IntToFloat(x), IntToFloat(y + Height, true), z, 0.0f, 0.0f,//bot l
                IntToFloat(x), IntToFloat(y, true), z, 0.0f, 1.0f// top l
            };
		}
		else
		{
			return new float[12] {
				IntToFloat(x + Width), IntToFloat(y, true), z,// top r
                IntToFloat(x + Width), IntToFloat(y + Height, true), z, //b r
                IntToFloat(x), IntToFloat(y + Height, true), z, //bot l
                IntToFloat(x), IntToFloat(y, true), z,// top l
            };
		}
	}

	public Vector3 PointToVector(float x, float y, float z = 0.0f)
	{
		return new Vector3(IntToFloat(x), IntToFloat(y, true), z);
	}

	public float IntToFloat(float p, bool Invert = false)
	{
		p += (Invert ? Location.Y : Location.X);
		IParent? tempp = Parent;
		while (tempp is not null)
		{
			p += (Invert ? tempp.Position.Y : tempp.Position.X);
			tempp = tempp.Parent;
		}
		float Size = (Invert ? Window!.Size.Y : Window!.Size.X);
		double half = Math.Round((double)Size / (double)2, 1);
		double Per = Math.Round((double)1 / half, 15);
		if (p == half) return 0.0f;
		if (Invert)
		{
			if (p > half) return (float)(((double)(p - half) * Per) * -1);
			else return (float)(1 - (p * Per));
		}
		else
		{
			if (p > half) return (float)((double)(p - half) * Per);
			else return (float)((1 - (p * Per)) * -1);
		}
	}

	public float FloatToInt(float p, bool Invert = false)
	{
		p += (Invert ? LocationAsFloat.Y : LocationAsFloat.X);
		IParent? tempp = Parent;
		while (tempp is not null)
		{
			p += (Invert ? tempp.LocationAsFloat.Y : tempp.LocationAsFloat.X);
			tempp = tempp.Parent;
		}

		float Size = (Invert ? Window!.Size.Y : Window!.Size.X);
		double half = Math.Round((double)Size / (double)2, 15);
		if (p == 0) return (int)half;
		if (Invert)
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(half * p);
			}
			else
			{
				return (float)(half - (p * half));
			}
		}
		else
		{
			if (p < 0)
			{
				p *= -1;
				p++;
				return (float)(Size - (half * p));
			}
			else
			{
				return (float)(p * half + half);
			}
		}
	}
	#endregion
}